var town_fshader_content = `
precision highp float;

/* Transforms that need to be applied */
uniform mat4 view;
uniform mat4 proj;

/* Lighting meta information (specular) */
uniform vec3 cameraPosition;
uniform float lightingProperties[8];

/* Specific light information */
uniform vec3 lightPositions[32]; /* Positions of the lights */
uniform vec4 lightColors[32]; /* Colors of each of the lights */
uniform float lightIntensities[32]; /* Light intensities */
uniform float lightFalloffs[32]; /* Light intensities */
uniform float lightTypes[32]; /* Type of light (disabled, ambient, point) */
uniform mat4 lightTx[32]; /* Transform for the lights */

/* Texture information */
uniform float useTexture;
// uniform sampler2D textureSample;

/* Material information */
uniform float hardness; /* Hardness of the object (specular) */

/* Stuff that must be passed from vertex shader */
varying vec3 fPosition;
varying vec3 fNormal;
// varying vec2 fTexCoord;
varying vec4 fDiffuse;
varying vec3 fCameraPosition;

void main()
{
    /* DEBUG */
    // vec4 c = vec4(fCameraPosition, 1.0);
    // c /= 20.0;
    // c.x = min(max(0.0, c.x), 1.0);
    // c.y = min(max(0.0, c.y), 1.0);
    // c.z = min(max(0.0, c.z), 1.0);
    // c.a = 1.0;
    // gl_FragColor = c;
    // return;

    // gl_FragColor = c;
    // return;
    // if(lightingProperties[5] == 0.0)
        // gl_FragColor = vec4(0.0, 1.0, 0.0, 1.0);
    // else gl_FragColor = vec4(0.0, 0.0, 0.0, 1.0);
    // return;
    //
    // gl_FragColor = vec4(abs(fNormal), 1.0);
    // return;

    float useDiffuse = lightingProperties[0];
    float useSpecular = lightingProperties[1];
    float useAmbient = lightingProperties[2];
    float useNormals = lightingProperties[3];
    float alphaAllowed = lightingProperties[4];
    float useAverage = lightingProperties[5];

    vec3 toCamera = normalize(fCameraPosition - fPosition);

    vec4 diffuseResult = vec4(0.0, 0.0, 0.0, 1.0);
    vec4 specularResult = vec4(0.0, 0.0, 0.0, 1.0);
    vec4 ambientResult = vec4(0.0, 0.0, 0.0, 1.0);

    int validLights = 0;
    for(int x = 0;x < 32;x++)
    {
        vec3 lightPosition = (proj * view * lightTx[x] * vec4(lightPositions[x], 1.0)).xyz;
        vec3 toLight = normalize(lightPosition - fPosition);

        vec3 v = normalize(toCamera);
        vec3 l = normalize(toLight);
        vec3 h = normalize(v + l);

        if(lightTypes[x] > 0.0)
        {
            validLights++;
            if(lightTypes[x] == 1.0)
            {
                /* Ambient Light */
                // ambientResult += lightColors[x] * lightIntensities[x];
                ambientResult += fDiffuse * lightIntensities[x];
            } else if(lightTypes[x] == 2.0 && useNormals > 0.0)
            {   
                vec4 diffuseColor = vec4(0.0, 0.0, 0.0, 1.0);

                if(useTexture > 100000.0)
                    return;
                // if(useTexture > 0.0)
                // {
                    // diffuseColor = texture2D(
                            // textureSample, fTexCoord).rgba; 
                // } else {
                    // diffuseColor = fDiffuse;
                // }
                
                diffuseColor = fDiffuse;

                /* Calculate the light falloff value */
                float d = distance(lightPosition, fPosition);
                if(d < 0.1)
                    d = 0.1;
                float fallI = 0.0;

                fallI = ((lightFalloffs[x] * 2.0) / d);

                if(fallI <= 0.0)
                    fallI = 0.0;
                else if(fallI > 1.0)
                    fallI = 1.0;

                fallI = 1.0;

                /* Point light */
                diffuseResult += 
                    diffuseColor
                    * fallI
                    * lightIntensities[x]
                    * max(0.0, dot(toLight, fNormal));
                    ;
                
                specularResult += 
                    lightColors[x]
                    * fallI
                    * (lightIntensities[x] * .3)
                    * pow(max(0.0, dot(fNormal, h)), hardness);
            }
        } else break;
    }

    /* Use only the values that are enabled */
    vec4 composite = vec4(0.0, 0.0, 0.0, 0.0);
    if(useAmbient > 0.0)
        composite += (ambientResult * 1.0);
    if(useDiffuse > 0.0)
        composite += diffuseResult;
    if(useSpecular > 0.0)
        composite += specularResult;

    /* Take the average of all lighting properties */
    // if(validLights > 0 && useAverage > 0.0)
        // composite /= float(validLights);

    composite.a = 1.0; /// hack?
    gl_FragColor = composite;
    // gl_FragColor = vec4(1.0, 0.0, 0.0, 1.0);
}
`;
