function WaveObj(object_text, name)
{
    this.debug = false;
    this.name = name;
    this.vertices = [];
    this.vertex_normals = [];
    this.faces = [];
    this.vertex_colors = [];
    this.tex_coords = [];
    this.gl_vertices = [];
    this.gl_normals = [];
    this.gl_texcoords = [];
    this.gl_diffuse = [];
    this.gl_center = [0, 0, 0];
    this.gl_texture = null;
    this.gl_hardness = 1.0;
    this.gl_rotate_tx = m4.identity();
    this.gl_translate_tx = m4.identity();
    this.gl_scale_tx = m4.identity();
    this.gl_tx = m4.identity();
    this.useGLTexture = false;
    this.globalColor = null;
    this.transform = null;
    this.animationCallback = null;

    var file_arr = object_text.split(/\r?\n/);

    var parsing = false;
    var failure = false;
    var objname = "";

    for(var x = 0;x < file_arr.length;x++)
    {
        var line = file_arr[x].trim();

        if(line == "") continue;
        if(line.indexOf("\#") == 0) continue;
        if(!parsing && line.indexOf("o") == 0)
        {
            parsing = true;
            objname = line.substr(1).trim();
            if(this.debug) console.log("Parsing new object: " + line.substr(1).trim());
            continue;
        } else if(parsing && line.indexOf("o") == 0)
        {
            alert("We don't support obj files with more than one object!");
            failure = true;
            break;
        }

        if(line.indexOf("vn") == 0)
        {
			// add a vertex
            var str = line.substr(2).trim().split(" ");
            var f0 = parseFloat(str[0]);
            var f1 = parseFloat(str[1]);
            var f2 = parseFloat(str[2]);
            this.vertex_normals.push([f0, f1, f2]);
            if(this.debug) console.log("Added vertex normal: " 
                    + str[0] + ", " + str[1] + ", " + str[2]);
            continue;
        }

        if(line.indexOf("vt") == 0)
        {
			// add a texture coordinate
            var str = line.substr(2).trim().split(" ");
            var f0 = parseFloat(str[0]);
            var f1 = parseFloat(str[1]);
            this.tex_coords.push([f0, f1]);
            if(this.debug) console.log("Added texture coordinate: " 
                    + str[0] + ", " + str[1]);
            continue;
        }

        if(line.indexOf("v") == 0)
        {
            // add a vertex
            var str = line.substr(1).trim().split(" ");
            var f0 = parseFloat(str[0]);
            var f1 = parseFloat(str[1]);
            var f2 = parseFloat(str[2]);
            this.vertices.push([f0, f1, f2]);
            if(this.debug) console.log("Added vertex: " + str[0] + ", " + str[1] + ", " + str[2]);
            continue;
        }

        if(line.indexOf("f") == 0)
        {
			// add a face
            var str = line.substr(1).trim().split(" ");
            this.processFace(str[0]);
            this.processFace(str[1]);
            this.processFace(str[2]);

            if(this.debug) console.log("Added face: " + str[0] + ", " + str[1] + ", " + str[2]);
            continue;
        }

        if(line.indexOf("c") == 0)
        {
            if(this.debug) console.log("Per vertex coloring not supported anymore.");
            // var str = line.substr(1).trim().split(" ");
            // console.log("Setting face " + (str[0] - 1) + " to color " + str[1]);
            // this.vertex_colors[parseInt(str[0]) - 1] = str[1];
            continue;
        }

        if(line.indexOf("C") == 0)
        {
            var str = line.split(" ");
            this.globalColor = this.parseColor(str[1].trim());
            this.gl_diffuse = this.getGLColors(1);
            if(this.debug) console.log("Color for entire object set to " + str[1]);
            continue;
        }

        if(this.debug) console.log("Warning unknown command: " + line);
    }

    if(failure)
    {
        this.vertices = [];
        this.faces = [];
        this.origVertices = [];
        this.gl_vertices = [];
        this.gl_normals = [];
        if(this.debug) console.log("Error: could not load object!!");
    } else {
        console.log("Object " + name + " added successfully.");
    }

    this.externalVertexBuffer = null;
    this.externalVertexBufferStart = -1;
    this.externalVertexBufferSize = -1;

    this.externalNormalBuffer = null;
    this.externalNormalBufferStart = -1;
    this.externalNormalBufferSize = -1;

    this.externalColorBuffer = null;
    this.externalColorBufferStart = -1;
    this.externalColorBufferSize = -1;

    this.shader_program = null;

    this.useGLVertices = false;
    this.useGLNormals = false;
    this.useGLDiffuse = false;
    this.useGLHardness = false;
}

/* make the graphics town happy */
WaveObj.prototype.init = function(){}


WaveObj.prototype.setShader = function(shader_program)
{
	this.shaderProgram = shader_program;
    return false;
}

WaveObj.prototype.initializeGLVertexBuffer = function(attrib_name)
{
    if(this.debug) console.log(this.name + ": Initializing vertex buffer: " + attrib_name);
    /* Get the attribute position */
    this.gl_vertices_attr = gl.getAttribLocation(this.shaderProgram, attrib_name);
    if(this.gl_vertices_attr < 0)
        return true;

    /* Create an uninitialized gl buffer */
    this.gl_vertices_buffer = gl.createBuffer();

    this.useGLVertices = true;
    if(this.debug) console.log(this.name + ":\tinitilize vertex buffer success.");
    return false;
}

WaveObj.prototype.initializeGLNormalBuffer = function(attrib_name)
{
    if(this.debug) console.log(this.name + ": Initializing normal buffer");
    /* Get the attribute position */
    this.gl_normals_attr = gl.getAttribLocation(this.shaderProgram, attrib_name);
    if(this.gl_normals_attr < 0)
        return true;

    /* Create an uninitialized gl buffer */
    this.gl_normals_buffer = gl.createBuffer();

    this.useGLNormals = true;
    if(this.debug) console.log(this.name + ": \tinitialize normal buffer success");
    return false;
}

WaveObj.prototype.initializeGLDiffuseBuffer = function(attrib_name)
{
    if(this.debug) console.log(this.name + ": Initializing diffuse buffer.");
    /* Get the attribute position */
    this.gl_diffuse_attr = gl.getAttribLocation(this.shaderProgram, attrib_name);
    if(this.gl_diffuse_attr < 0)
        return true;

    /* Create an uninitialized gl buffer */
    this.gl_diffuse_buffer = gl.createBuffer();

    this.useGLDiffuse = true;
    if(this.debug) console.log(this.name + ": \tinitializing diffuse buffer success");
    return false;;
}

WaveObj.prototype.initializeGLTextureBuffer = function(useTexture, textureSample, texCoords)
{
    this.useGLTexture = false;
    if(this.debug) console.log(this.name + ": Texture buffer is disabled.");
    this.gl_use_texture_uniform = 
        gl.getUniformLocation(this.shaderProgram, useTexture);
    // this.gl_texture_sample_uniform =
        // gl.getUniformLocation(this.shaderProgram, textureSample);

    /* TODO: enable textures here */
    return this.gl_use_texture_uniform == null;
}

WaveObj.prototype.initializeGLHardness = function(uniform_name, hardness)
{
    this.useGLHardness = true;
    this.gl_hardness = hardness;
    this.gl_hardness_uniform = 
        gl.getUniformLocation(this.shaderProgram, uniform_name);
    return this.gl_hardness_uniform == null;
}

WaveObj.prototype.initializeGLTx = function(uniform_name)
{
    this.gl_tx_uniform = 
        gl.getUniformLocation(this.shaderProgram, uniform_name);
    return this.gl_tx_uniform == null;
}

WaveObj.prototype.setHardness = function(hardness)
{
    this.gl_hardness = hardness;
}

WaveObj.prototype.setTexture = function(source)
{
    this.gl_texture = gl.createTexture();
    this.texturLoaded = false;
    var image = new Image();
    image.src = source;

    image.onload = function() 
    { 
  		gl.bindTexture(gl.TEXTURE_2D, this.gl_texture);
  		gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, image);
  		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
  		gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR_MIPMAP_NEAREST);
  		gl.generateMipmap(gl.TEXTURE_2D);
  		gl.bindTexture(gl.TEXTURE_2D, null);
		this.textureLoaded = true;
    };
}

WaveObj.prototype.parseColor = function(s)
{
	var r = 0;
    var g = 0;
    var b = 0;
    var a = 3;
    switch(s.toLowerCase())
    {
        case "grey":
            r = 200;
            g = 200;
            b = 200;
            break;
        case "red":
            r = 128;
            g = 0;
            b = 0;
            break;
        case "orange":
            r = 128;
            g = 85;
            b = 0;
            break;
        case "yellow":
            r = 160;
            g = 160;
            b = 0;
            break;
        case "green":
            r = 0;
            g = 120;
            b = 0;
            break;
        case "blue":
            r = 0;
            g = 0;
            b = 140;
            break;
        case "purple":
            r = 128;
            g = 0;
            b = 128;
            break;
        case "white":
            r = 170;
            g = 170;
            b = 170;
            break;
        case "black":
            r = 0;
            g = 0;
            b = 0;
            break;
        case "brown":
            r = 80;
            g = 50;
            b = 50;
            break;
        default: /* Convert hex value */
            if(s.length == 9)
            {
                r = parseInt(s.substr(1, 2), 16);
                g = parseInt(s.substr(3, 2), 16);
                b = parseInt(s.substr(5, 2), 16);
                a = parseInt(s.substr(7, 2), 16);
            } else {
                console.log("Error: invalid color: " + s);
            }
            break;
    }

	return [r, g, b, a];
}

WaveObj.prototype.center = function()
{
    return m4.transformPoint(this.gl_tx, this.gl_center);
}

WaveObj.prototype.setCenter = function(c)
{
    if(c.length != 3)
        return;
    this.gl_center = c;
}

WaveObj.prototype.getGLTx = function()
{
    return this.gl_tx;
}

WaveObj.prototype.updateGLTx = function()
{
    this.gl_tx = m4.multiply(
            m4.multiply(this.gl_rotate_tx, this.gl_scale_tx),
            this.gl_translate_tx);
}

/**
 * Reset all transformations
 */
WaveObj.prototype.resetTx = function()
{
    this.gl_scale_tx = m4.identity();
    this.gl_translate_tx = m4.identity();
    this.gl_rotate_tx = m4.identity;
    this.gl_tx = m4.identity();
}

/**
 * Rotate the object. rot is a vector of 3 numbers;
 */
WaveObj.prototype.rotate = function(rot)
{
    var x = m4.rotationX(rot[0]);
    var y = m4.rotationY(rot[1]);
    var z = m4.rotationZ(rot[2]);
    var newrot = m4.multiply(z, m4.multiply(y, x));

    this.gl_rotate_tx = m4.multiply(newrot, this.gl_rotate_tx);
    this.updateGLTx();
}

WaveObj.prototype.setRotation = function(rot)
{
    var x = m4.rotationX(rot[0]);
    var y = m4.rotationY(rot[1]);
    var z = m4.rotationZ(rot[2]);

    this.gl_rotate_tx = m4.multiply(z, m4.multiply(y, x));
    this.updateGLTx();
}

/**
 * Translate the object. loc is a vector of 3 numbers.
 */
WaveObj.prototype.translate = function(loc)
{
    this.gl_translate_tx = m4.multiply(loc, this.gl_translate_tx);
    this.updateGLTx();
}

WaveObj.prototype.setTranslation = function(tx)
{
    this.gl_translate_tx = m4.translation(tx);
    console.log("Translationx: " + this.gl_translate_tx);
    this.updateGLTx();
}

WaveObj.prototype.processFace = function(faceStr)
{
    if(faceStr.indexOf("//") > 0)
    {
        var line = faceStr.split("//");
        var v = parseInt(line[0]) - 1;
        var vn = parseInt(line[1]) - 1;

        var vertex = this.vertices[v];
        var normal = this.vertex_normals[vn];

        for(var i = 0;i < vertex.length;i++)
            this.gl_vertices.push(vertex[i]);
        for(var i = 0;i < vertex.length;i++)
            this.gl_normals.push(normal[i]);


        return false;
    } else if(faceStr.indexOf("/") > 0)
    {
        var parts = faceStr.trim().split("/");

        if(parts.length == 2)
        {
            /* Only vertex / tex coord */
            var vertexn = parseInt(parts[0]) - 1;
            var tex_coordn = parseInt(parts[1]) - 1;

            var vertex = this.vertices[vertexn];
            var tex_coord = this.tex_coords[tex_coordn];

            for(var i = 0;i < vertex.length;i++)
                this.gl_vertices.push(vertex[i]);
            for(var i = 0;i < tex_coord.length;i++)
                this.gl_texcoords.push(tex_coord[i]);
            return false;

        } else if(parts.length == 3)
        {
            /* Vertex / Texture coord / Vertex Normal */
            var vertexn = parseInt(parts[0]) - 1;
            var tex_coordn = parseInt(parts[1]) - 1;
            var normaln = parseInt(parts[2]) - 1;

            var vertex = this.vertices[vertexn];
            var tex_coord = this.tex_coords[tex_coordn];
            var normal = this.vertex_normals[normaln];

            for(var i = 0;i < vertex.length;i++)
                this.gl_vertices.push(vertex[i]);
            for(var i = 0;i < tex_coord.length;i++)
                this.gl_texcoords.push(tex_coord[i]);
            for(var i = 0;i < normal.length;i++)
                this.gl_normals.push(normal[i]);
            return false;
        } else {
            if(this.debug) console.log("Unsupported face format: " + faceStr);
        }
    } else {
        var v = parseInt(faceStr) - 1;
        var vertex = this.vertices[v];
        for(var i = 0;i < vertex.length;i++)
            this.gl_vertices.push(vertex[i]);
        return false;
    }

    return true;
}

/**
 * Make the entire object a single color
 */
WaveObj.prototype.setGlobalColor = function(c, normalize)
{
    this.globalColor = this.parseColor(c);
    this.gl_color = this.getGLColors(normalize);
    this.gl_diffuse = this.gl_color;

    if(this.externalColorBuffer)
    {
        var buffer = this.externalColorBuffer;
        var start = this.externalColorBufferStart;
        for(var i = 0;i < this.externalColorBufferSize;i += 4)
        {
            buffer[start + i + 0] = this.globalColor[0] * normalize;
            buffer[start + i + 1] = this.globalColor[1] * normalize;
            buffer[start + i + 2] = this.globalColor[2] * normalize;
            buffer[start + i + 3] = this.globalColor[3];
        }
    }
}

WaveObj.prototype.getGLVertices = function()
{
    return this.gl_vertices;
}

WaveObj.prototype.getGLNormals = function()
{
    return this.gl_normals;
}

WaveObj.prototype.getPrimitiveCount = function()
{
    return this.gl_vertices.length / 3;
}

WaveObj.prototype.getGLColors = function(normalize)
{
    if(this.globalColor)
    {
        if(normalize > 0)
        {
            var arr = [];
            for(var i = 0;i < this.getPrimitiveCount();i++)
            {
                arr.push(this.globalColor[0] / normalize);
                arr.push(this.globalColor[1] / normalize);
                arr.push(this.globalColor[2] / normalize);
                arr.push(this.globalColor[3]);
            }


            return arr;
        } else conole.log("ERROR: Normalize must be > 0: " + normalize);
    } else console.log("Error: No colors for this object!!");

    return null;
}

WaveObj.prototype.draw = function()
{
    gl.useProgram(this.shaderProgram);

    if(this.useGLVertices)
    {
        /* Setup the vertices attribute */
        gl.enableVertexAttribArray(this.gl_vertices_attr);
        gl.bindBuffer(gl.ARRAY_BUFFER, this.gl_vertices_buffer);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(this.gl_vertices), gl.STATIC_DRAW);
        gl.vertexAttribPointer(this.gl_vertices_attr, 3, gl.FLOAT, false, 0, 0);

        if(this.useGLNormals)
        {
            /* Setup the normals attribute */
            gl.enableVertexAttribArray(this.gl_normals_attr);
            gl.bindBuffer(gl.ARRAY_BUFFER, this.gl_normals_buffer);
            gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(this.gl_normals), gl.STATIC_DRAW);
            gl.vertexAttribPointer(this.gl_normals_attr, 3, gl.FLOAT, false, 0, 0);
        } else if(this.debug) console.log("WARNING: no normal data supplied. ");

        if(this.useGLDiffuse)
        {
            /* Setup the normals attribute */
            gl.enableVertexAttribArray(this.gl_diffuse_attr);
            gl.bindBuffer(gl.ARRAY_BUFFER, this.gl_diffuse_buffer);
            gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(this.gl_diffuse), gl.STATIC_DRAW);
            gl.vertexAttribPointer(this.gl_diffuse_attr, 4, gl.FLOAT, false, 0, 0);
        } else if(this.debug) console.log("WARNING: no diffuse data supplied.");

        /* Should we use texture? */
        if(this.useGLTexture)
        {
             
        } else {
            gl.uniform1f(this.gl_use_texture_uniform, 0.0);
        }

        if(this.useGLHardness)
        {
            gl.uniform1f(this.gl_hardness_uniform, this.gl_hardness);
        } else console.log("ERROR: no hardness availble for " + this.name);

        /* Set our transformation */
        if(this.gl_tx_uniform != null)
        {
            gl.uniformMatrix4fv(this.gl_tx_uniform, false, this.gl_tx);
        } else console.log("Cannot set model TX!!");

        /* Do the draw */
        gl.drawArrays(gl.TRIANGLES, 0, this.gl_vertices.length / 3);

    } else console.log("ERROR: No vertex data!  " + this.verticesAttr);
}

WaveObj.prototype.setAnimationCallback = function(f)
{
    this.animationCallback = f;
}

WaveObj.prototype.animate = function(args)
{
    if(this.animationCallback != null)
        this.animationCallback(this, args);
}
