var debug = true;

function Scene(vshader_content, fshader_content)
{
    if(debug) console.log("Initializing scene...");
    this.vshader_content = vshader_content;
    this.fshader_content = fshader_content;
    this.objects = [];

    this.vshader = null;
    this.fshader = null;
    this.shaderProgram = null;

    this.verticesAttr = -1;
    this.verticesBuffer = null;
    this.primitiveCount = -1;

    this.normalsAttr = -1;
    this.normalsBuffer = null;

    this.colorsAttr = -1;
    this.colorsBuffer = null;

    this.worldTx = null;
    this.perspectiveTx = null;
    this.cameraUniform = null;

    this.updateWorldTx();
    this.colorNormalization = 255;
    if(debug) console.log("\tColor normalization: " + this.colorNormalization);

    this.rotationX = 0;
    this.rotationY = 0;

    this.cameraFrom = [0.0, 0.0, 0.0];

    this.useGLLighting = false;
    this.gl_use_transforms = false;
    this.gl_have_attrib_names = false;
    this.initLightBuffers();
}

Scene.prototype.initBuffer = function(buffer, count)
{
    for(var i = 0;i < count;i++)
        buffer.push(0.0);
}

Scene.prototype.initLightBuffers = function()
{
    this.lightCount = 32;
    this.lightingProperties = [];
    this.initBuffer(this.lightingProperties, 8);

    this.lightPositions = [];
    this.lightColors = [];
    this.lightIntensities = [];
    this.lightFalloffs = [];
    this.lightTypes = [];
    this.lightTxs = [];
    this.lightAnimationCallbacks = [];

    this.initBuffer(this.lightPositions, this.lightCount * 3);
    this.initBuffer(this.lightColors, this.lightCount * 4);
    this.initBuffer(this.lightIntensities, this.lightCount * 1);
    this.initBuffer(this.lightFalloffs, this.lightCount * 1);
    this.initBuffer(this.lightTypes, this.lightCount * 1);
    this.initBuffer(this.lightTxs, this.lightCount * 16);

    for(var i = 0;i < this.lightCount;i++)
        this.lightAnimationCallbacks.push(null);

    this.enableDiffuse();
    this.enableSpecular();
    this.enableAmbient();
    this.enableNormals();

    if(debug) console.log("\tLight buffers initialized.");
    if(debug) console.log("\tLight properties: " + this.lightingProperties);
}

/** Uniform functions */

Scene.prototype.initGLLighting = function(property_uname, position_uname,
        color_uname, intensity_uname, falloff_uname, type_uname, cposition_uname,
        tx_uname)
{
    if(debug) console.log("\tInitializing light uniforms...");
    this.gl_lighting_properties_uniform = 
        gl.getUniformLocation(this.shaderProgram, property_uname);
    this.gl_light_position_uniform = gl.getUniformLocation(this.shaderProgram, position_uname);
    this.gl_light_color_uniform = gl.getUniformLocation(this.shaderProgram, color_uname);
    this.gl_light_intensity_uniform = gl.getUniformLocation(this.shaderProgram, intensity_uname);
    this.gl_light_falloff_uniform = gl.getUniformLocation(this.shaderProgram, falloff_uname);
    this.gl_light_type_uniform = gl.getUniformLocation(this.shaderProgram, type_uname);
    this.gl_light_tx_uniform = gl.getUniformLocation(this.shaderProgram, tx_uname);

    if(this.gl_lighting_properties_uniform == null)
    {
        console.log("ERROR: Light properties uniform is null!");
        return true;
    }

    if(this.gl_light_position_uniform == null)
    {
        console.log("ERROR: Light position uniform is null!");
        return true;
    }

    if(this.gl_light_color_uniform == null)
    {
        console.log("ERROR: Light color uniform is null!");
        return true;
    }

    if(this.gl_light_intensity_uniform == null)
    {
        console.log("ERROR: Light intensity uniform is null!");
        return true;
    }

    if(this.gl_light_falloff_uniform == null)
    {
        console.log("ERROR: Light falloff uniform is null!");
        return true;
    }

    if(this.gl_light_type_uniform == null)
    {
        console.log("ERROR: Light type uniform is null!");
        return true;
    }

    if(this.gl_light_tx_uniform == null)
    {
        console.log("ERROR: Light TX uniform is null!");
        return true;
    }

    this.useGLLighting = true;

    if(debug) console.log("\tScene GLLighting appears to be okay.");

    return false;
}

Scene.prototype.setView = function(viewTx, projTx, cameraPos)
{
    if(cameraPos.length < 3)
        return true;;

    this.gl_view = viewTx;
    this.gl_proj = projTx;
    this.gl_camera_position = cameraPos;

    return false;
}

Scene.prototype.setViewUniforms = function(viewTx, projTx, camera)
{
    this.gl_view_uniform = gl.getUniformLocation(this.shaderProgram, viewTx);
    this.gl_proj_uniform = gl.getUniformLocation(this.shaderProgram, projTx);
    this.gl_camera_position_uniform = gl.getUniformLocation(this.shaderProgram, camera);
    return this.gl_view_uniform == null 
        || this.gl_proj_uniform == null
        || this.gl_camera_position_uniform == null;
}

Scene.prototype.setTextureUniforms = function(useTexture, textureSample, hardness)
{
    this.gl_use_texture = useTexture;
    this.gl_texture_sample = textureSample;
    this.gl_hardness = hardness;

    return false;
}

Scene.prototype.verifyGLUniforms = function()
{
    if(this.gl_view_uniform == null
            || this.gl_proj_uniform == null
            || this.gl_camera_position_uniform == null
            || this.gl_lighting_properties_uniform == null
            || this.gl_light_position_uniform == null
            || this.gl_light_color_uniform == null
            || this.gl_light_falloff_uniform == null
            || this.gl_light_intensity_uniform == null
            || this.gl_light_type_uniform == null)
    {
        console.log("\t\tREQUIRED UNIFORMS NOT SET: "
            + "V: " + this.gl_view_uniform
            + "P: " +  this.gl_proj_uniform
            + "CP: " +  this.gl_camera_position_uniform
            + "LPROP: " +  this.gl_lighting_properties_uniform
            + "LPOS: " +  this.gl_light_position_uniform
            + "LC: " +  this.gl_light_color_uniform
            + "LI: " +  this.gl_light_intensity_uniform
            + "LF: " +  this.gl_light_falloff_uniform
            + "LT: " +  this.gl_light_type_uniform);
        return true;
    }

    return false;
}

Scene.prototype.setAttributeNames = function(vertices, normals, diffuse, texCoords, gltx)
{
    this.gl_vertices = vertices;
    this.gl_normals = normals;
    this.gl_diffuse = diffuse;
    this.gl_texture_coords = texCoords;
    this.gl_tx = gltx;
    this.gl_have_attrib_names = true;
}

Scene.prototype.updateLightTx = function(index, Tx)
{
    if(index < this.lightCount && index >= 0)
    {
        var start = index * 16;
        for(var i = 0; i < 16; i++)
        {
            this.lightTxs[start + i] = Tx[i];
        }
    }
}

Scene.prototype.getLightTx = function(index)
{
    if(index < this.lightCount && index >= 0)
        return this.lightTxs[index];
    return null;
}

Scene.prototype.setLightAnimationCallback = function(index, callback)
{
    if(index < this.lightCount && index >= 0)
        this.lightAnimationCallbacks[index] = callback;
}

/**
 * Add a light to the scene 
 */
Scene.prototype.addLight = function(position, color, intensity, falloff, typestr)
{
    if(position.length < 3)
        return true;
    if(color.length < 4)
        return true;

    if(debug) console.log("\tAdding a light to the scene...");
    var type = 0;
    if(typestr == "ambient")
        type = 1;
    else if(typestr == "point")
        type = 2;
    else return true;

    /* Find an unused light */
    var index = 0;
    for(;index < this.lightCount;index++)
    {
        if(this.lightTypes[index] <= 0.0)
            break;
    }

    /* Did we find one? */
    if(index < this.lightCount)
    {
        var pIndex = index * 3;
        var cIndex = index * 4;

        this.lightPositions[pIndex + 0] = position[0];
        this.lightPositions[pIndex + 1] = position[1];
        this.lightPositions[pIndex + 2] = position[2];

        this.lightColors[cIndex + 0] = color[0];
        this.lightColors[cIndex + 1] = color[1];
        this.lightColors[cIndex + 2] = color[2];
        this.lightColors[cIndex + 3] = color[3];

        this.lightIntensities[index] = intensity;
        this.lightFalloffs[index] = falloff;
        this.lightTypes[index] = type;
        this.updateLightTx(index, m4.identity());

    } else return true;

    if(false)
    {
    console.log("positions  : " + this.lightPositions);
    console.log("colors     : " + this.lightColors);
    console.log("intensities: " + this.lightIntensities);
    console.log("falloffs   : " + this.lightFalloffs);
    console.log("types      : " + this.lightTypes);
    }

    if(debug) console.log("\tLight added successfully.");
    return false;
}

/* Enable/Disable light effects */
Scene.prototype.enableDiffuse = function() { this.lightingProperties[0] = 1.0; }
Scene.prototype.disableDiffuse = function() { this.lightingProperties[0] = 0.0; }
Scene.prototype.enableSpecular = function() { this.lightingProperties[1] = 1.0; }
Scene.prototype.disableSpecular = function() { this.lightingProperties[1] = 0.0; }
Scene.prototype.enableAmbient = function() { this.lightingProperties[2] = 1.0; }
Scene.prototype.disableAmbient = function() { this.lightingProperties[2] = 0.0; }
Scene.prototype.enableNormals = function() { this.lightingProperties[3] = 1.0; }
Scene.prototype.disableNormals = function() { this.lightingProperties[3] = 0.0; }
Scene.prototype.allowAlpha = function() { this.lightingProperties[4] = 1.0; }
Scene.prototype.denyAlpha = function() { this.lightingProperties[4] = 0.0; }
Scene.prototype.enableLightAverage = function() { this.lightingProperties[5] = 1.0; }
Scene.prototype.disableLightAverage = function() { this.lightingProperties[5] = 0.0; }

/**
 * Compile a glsl shader (internal use only)
 */
function compileShader(gl, type, source)
{
    /* Have WebGL create the shader for us */
    console.log("Creating shader of type " + type);
    var shader = gl.createShader(type);
    /* Set the shader source */
    gl.shaderSource(shader, source);

    /* Compile the shader */
    console.log("\tCompiling shader...");
    gl.compileShader(shader);

    /* Check the compile status */
    if(gl.getShaderParameter(shader, gl.COMPILE_STATUS))
    {
        console.log("\tCompile success.");
        return shader;
    }

    /* Compiling failed */
    console.log("\tCompiler error.");
   
    /* Print any compiler output */
    console.log(gl.getShaderInfoLog(shader));
    gl.deleteShader(shader);
    return null;
}

/**
 * Create a glsl program from a vertex and fragment shader (internal use only)
 */
function createProgram(gl, vertexShader, fragmentShader) 
{
    /* Create a shader program */
    console.log("Creating shader program...");
    var program = gl.createProgram();

    /* Attach both shaders */
    console.log("\tAttaching shaders...");
    gl.attachShader(program, vertexShader);
    gl.attachShader(program, fragmentShader);

    /* Link shaders */
    console.log("\tLinking shaders...");
    gl.linkProgram(program);

    /* Check link status */
    if(gl.getProgramParameter(program, gl.LINK_STATUS))
    {
        console.log("\tLink success!");
        return program;
    }

    /* Link failure */
    console.log("\tLink failure.");
    console.log(gl.getProgramInfoLog(program));
    gl.deleteProgram(program);
    return null;
}

/* Compile the shaders given to use during initialization */
Scene.prototype.compileShaders = function()
{
    this.vshader = compileShader(gl, gl.VERTEX_SHADER, this.vshader_content);

    if(!this.vshader)
        return true;

    this.fshader = compileShader(gl, gl.FRAGMENT_SHADER, this.fshader_content);

    if(!this.fshader)
    {
        this.vshader = null;
        return true;
    }

    return false;
}

/* Create the shader program */
Scene.prototype.linkShaders = function()
{
    this.shaderProgram = createProgram(gl, this.vshader, this.fshader);
    if(!this.shaderProgram)
        return true;

    return false;
}


Scene.prototype.setupView = function(rx, ry)
{
    this.rx = rx;
    this.ry = ry;
    console.log("View resolution is " + rx + "x" + ry);
}

Scene.prototype.addObject = function(wave)
{
    /* Use our global shader */
    wave.setShader(this.shaderProgram);

    /* Setup properties */
    if(wave.initializeGLVertexBuffer(this.gl_vertices))
    {
        console.log("FAILURE: Object couldn't setup its vertex buffer!");
        return true;
    }

    if(wave.initializeGLNormalBuffer(this.gl_normals))
    {
        console.log("FAILURE: Object couldn't setup its normals buffer!");
        return true;
    }

    if(wave.initializeGLDiffuseBuffer(this.gl_diffuse))
    {
        console.log("FAILURE: Object couldn't setup its diffuse buffer!");
        return true;
    }

    if(wave.initializeGLTextureBuffer(this.gl_use_texture, 
                this.gl_texture_sample, this.gl_texture_coords))
    {
        console.log("FAILURE: Object couldn't setup its texture buffer!");
        return true;
    }

    if(wave.initializeGLHardness(this.gl_hardness, 2.0))
    {
        console.log("FAILURE: Object couldn't setup its hardness uniform!");
        return true;
    }

    if(wave.initializeGLTx(this.gl_tx))
    {
        console.log("FAILURE: Object couldn't setup its model TX uniform!");
        return true;
    }

    this.objects.push(wave);

    return false;
}

Scene.prototype.lookAt = function(from, to, up)
{
    this.cameraFrom = from.slice();
    this.cameraTo = to.slice();
    this.cameraUp = up.slice();

    this.cameraTx = m4.inverse(m4.lookAt(this.cameraFrom, this.cameraTo, this.cameraUp));
    this.updateWorldTx();
}

Scene.prototype.setPerspective = function(perspective)
{
    if(perspective == true)
        this.perspectiveTx = this.perspectiveTxSave;
    else this.perspectiveTx = null;
    this.updateWorldTx();
}

Scene.prototype.enablePerspective = function(fov, aspect, z1, z2)
{
    console.log("Enabling perspective transform...");
    console.log("\tField of view: " + fov + "  in radians: " 
            + (fov / 180 * Math.PI));
    console.log("\tAspect ratio: " + aspect + ":1");
    console.log("\tClipping start: " + z1 + " clipping end: " + z2);
    this.perspectiveTx = m4.perspective(fov / 180 * Math.PI, aspect, z1, z2);
    this.perspectiveTxSave = this.perspectiveTx;
    this.updateWorldTx();
}

Scene.prototype.enableOrthographic = function()
{
    this.perspectiveTx = null;
    this.updateWorldTx();
}

Scene.prototype.rotateX = function(val)
{
    this.rotationX += val;
    this.updateWorldTx();
}

Scene.prototype.rotateY = function(val)
{
    this.rotationY += val;
    this.updateWorldTx();
}

Scene.prototype.updateWorldTx = function()
{
    if(this.cameraTx == null)
    {
        this.lookAt([0, 0, -10], [0, 0, 0], [0, 1, 0]);
        return;
    }

	var wtx = null;
    /* Apply projection if we are in perspective mode */
    if(this.perspectiveTx)
        wtx = m4.multiply(this.cameraTx, this.perspectiveTx);
    else {
        // var ortho = m4.ortho(-this.rx / 2, this.rx / 2, this.ry / 2, -this.ry / 2,
                // 1, 10000);
        var ortho = m4.ortho(-this.rx, this.rx, -this.ry, this.ry,
                1, 10000);
        wtx = m4.multiply(this.cameraTx, ortho);
    }

    var rotX = m4.rotationX(this.rotationX);
    var rotFinal = m4.rotateY(rotX, this.rotationY);

    // wtx = m4.inverse(wtx);

    /* Apply all viewport transformations */
    this.worldTx = m4.multiply(rotFinal, wtx); //m4.transpose(wtx);
}


Scene.prototype.setCameraPosition = function(pos)
{
    this.cameraFrom = pos.slice();
}

Scene.prototype.animate = function(args)
{
    for(var i = 0;i < this.objects.length;i++)
        this.objects[i].animate(args);

    for(var i = 0;i < this.lightCount;i++)
        if(this.lightAnimationCallbacks[i] != null)
        {
            this.lightAnimationCallbacks[i](this, i, args);
        }
}

Scene.prototype.draw = function(examine)
{
    // gl.viewport(0, 0, this.rx, this.ry);
    // gl.clearColor(0, 0, 0, 0);

    /* Clear color and depth buffers */
    // gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);

    /* Enable face culling (performance) */
    // gl.enable(gl.CULL_FACE);
    // gl.enable(gl.DEPTH_TEST);

    /* Tell WebGL to use our shader */
    gl.useProgram(this.shaderProgram);

    // if(this.verticesAttr >= 0)
    // {
        // /* Setup the vertices attribute */
        // gl.enableVertexAttribArray(this.verticesAttr);
        // gl.bindBuffer(gl.ARRAY_BUFFER, this.verticesBuffer);
        // gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(this.getGLVertices()), 
            // gl.STATIC_DRAW);
        // gl.vertexAttribPointer(this.verticesAttr, 3, gl.FLOAT, false, 0, 0);
    // } else console.log("ERROR: No vertex data!  " + this.verticesAttr);

    // if(this.normalsAttr >= 0)
    // {
        // gl.enableVertexAttribArray(this.normalsAttr);
        // gl.bindBuffer(gl.ARRAY_BUFFER, this.normalsBuffer);
        // gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(this.getGLNormals()),
            // gl.STATIC_DRAW);
        // gl.vertexAttribPointer(this.normalsAttr, 3, gl.FLOAT, false, 0, 0);
    // }

    // if(this.colorsAttr >= 0)
    // {
        // gl.enableVertexAttribArray(this.colorsAttr);
        // gl.bindBuffer(gl.ARRAY_BUFFER, this.colorsBuffer);
        // gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(this.getGLColors()),
                // gl.STATIC_DRAW);
        // gl.vertexAttribPointer(this.colorsAttr, 4, gl.FLOAT, false, 0, 0);
    // }

    // if(!this.worldTx)
    // {
        // console.log("WorldTX never initialized!");
    // }

    /* Apply the world transform */
    // if(this.worldTxUniform)
    // {
        // gl.uniformMatrix4fv(this.worldTxUniform, false, this.worldTx);
    // } else console.log("ERROR: No world transform!");

    /* Apply the lighting */

    if(!this.useGLLighting)
    {
        console.log("GL LIGHTING NOT SETUP.");
        return true;
    }

    /* Set all of the lighting uniforms */
    gl.uniform1fv(this.gl_lighting_properties_uniform, 
            new Float32Array(this.lightingProperties));
    gl.uniform3fv(this.gl_light_position_uniform, 
            new Float32Array(this.lightPositions));
    gl.uniform4fv(this.gl_light_color_uniform, 
            new Float32Array(this.lightColors));
    gl.uniform1fv(this.gl_light_intensity_uniform, 
            new Float32Array(this.lightIntensities));
    gl.uniform1fv(this.gl_light_falloff_uniform, 
            new Float32Array(this.lightFalloffs));
    gl.uniform1fv(this.gl_light_type_uniform, 
            new Float32Array(this.lightTypes));
    gl.uniformMatrix4fv(this.gl_light_tx_uniform, false, this.lightTxs);

    /* Set view information*/
    gl.uniform3fv(this.gl_camera_position_uniform, 
            new Float32Array(this.gl_camera_position));
    gl.uniformMatrix4fv(this.gl_view_uniform, false, this.gl_view);
    gl.uniformMatrix4fv(this.gl_proj_uniform, false, this.gl_proj);

    /* Draw the sub components */
    for(var i = 0;i < this.objects.length && !examine;i++)
    {
        this.objects[i].draw();    
    }
}

